use rand::{thread_rng, Rng};
use std::cmp::{Ord, Ordering, PartialOrd};
use std::fmt;
use std::str::FromStr;

/// Weapons for Rock paper scissors
#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
pub enum Weapon {
    /// A rock, it beats scissors
    Rock,
    /// Paper, it beats a rock
    Paper,
    /// Scissors, it beats paper
    Scissors,
}

/// Error parsing weapon
#[derive(Debug, Fail, Clone, Copy)]
pub enum WeaponParseError {
    /// Weapon attempted to be parsed was not valid
    #[fail(display = "Invalid Weapon")]
    InvalidWeapon,
}

impl fmt::Display for Weapon {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Weapon::Rock => "Rock",
                Weapon::Paper => "Paper",
                Weapon::Scissors => "Scissors",
            }
        )
    }
}

impl FromStr for Weapon {
    type Err = WeaponParseError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        s.chars()
            .nth(0)
            .and_then(|c| Some(c.to_ascii_lowercase()))
            .and_then(|c| match c {
                'r' => Some(Weapon::Rock),
                'p' => Some(Weapon::Paper),
                's' => Some(Weapon::Scissors),
                _ => None,
            })
            .ok_or(WeaponParseError::InvalidWeapon)
    }
}

impl PartialOrd for Weapon {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for Weapon {
    fn cmp(&self, other: &Self) -> Ordering {
        match self {
            Weapon::Rock => match other {
                Weapon::Rock => Ordering::Equal,
                Weapon::Paper => Ordering::Less,
                Weapon::Scissors => Ordering::Greater,
            },
            Weapon::Paper => match other {
                Weapon::Rock => Ordering::Greater,
                Weapon::Paper => Ordering::Equal,
                Weapon::Scissors => Ordering::Less,
            },
            Weapon::Scissors => match other {
                Weapon::Rock => Ordering::Less,
                Weapon::Paper => Ordering::Greater,
                Weapon::Scissors => Ordering::Equal,
            },
        }
    }
}

impl Weapon {
    /// Randomly choose a weapon
    pub fn rand() -> Self {
        "rps"
            .chars()
            .nth(thread_rng().gen_range(0, 3))
            .and_then(|p| Weapon::from_str(&p.to_string()).ok())
            .expect("Rand should always return weapon")
    }
}
